@@ -1,4 +1,4 @@
-/* Copyright (C) 2010, 2017, MariaDB Corporation Ab
+/* Copyright (C) 2010, 2020, 2021, MariaDB Corporation.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +11,7 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02111-1301 USA */
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1335 USA */
 
 /**
   @file
@@ -21,11 +21,14 @@
 
 #include "mariadb.h"
 #include "create_options.h"
+#include "partition_info.h"
 #include <my_getopt.h>
 #include "set_var.h"
 
 #define FRM_QUOTED_VALUE 0x8000U
 
+static const char *bools="NO,OFF,0,YES,ON,1";
+
 /**
   Links this item to the given list end
 
@@ -43,15 +46,12 @@ void engine_option_value::link(engine_option_value **start,
   engine_option_value *opt;
   /* check duplicates to avoid writing them to frm*/
   for(opt= *start;
-      opt && ((opt->parsed && !opt->value.str) ||
-              my_strnncoll(system_charset_info,
-                           (uchar *)name.str, name.length,
-                           (uchar*)opt->name.str, opt->name.length));
+      opt && ((opt->parsed && !opt->value.str) || !name.streq(opt->name));
       opt= opt->next) /* no-op */;
   if (opt)
   {
-    opt->value.str= NULL;       /* remove previous value */
-    opt->parsed= TRUE;          /* and don't issue warnings for it anymore */
+    opt->value= Value(); /* remove previous value */
+    opt->parsed= TRUE;   /* and don't issue warnings for it anymore */
   }
   /*
     Add this option to the end of the list
@@ -98,14 +98,13 @@ static bool report_unknown_option(THD *thd, engine_option_value *val,
 {
   DBUG_ENTER("report_unknown_option");
 
-  if (val->parsed || suppress_warning)
+  if (val->parsed || suppress_warning || thd->slave_thread)
   {
     DBUG_PRINT("info", ("parsed => exiting"));
     DBUG_RETURN(FALSE);
   }
 
-  if (!(thd->variables.sql_mode & MODE_IGNORE_BAD_TABLE_OPTIONS) &&
-      !thd->slave_thread)
+  if (!(thd->variables.sql_mode & MODE_IGNORE_BAD_TABLE_OPTIONS))
   {
     my_error(ER_UNKNOWN_OPTION, MYF(0), val->name.str);
     DBUG_RETURN(TRUE);
@@ -119,20 +118,19 @@ static bool report_unknown_option(THD *thd, engine_option_value *val,
 
 #define value_ptr(STRUCT,OPT)    ((char*)(STRUCT) + (OPT)->offset)
 
-static bool set_one_value(ha_create_table_option *opt,
-                          THD *thd, const LEX_CSTRING *value, void *base,
-                          bool suppress_warning,
-                          MEM_ROOT *root)
+static bool set_one_value(ha_create_table_option *opt, THD *thd,
+                          const engine_option_value::Value *value, void *base,
+                          bool suppress_warning, MEM_ROOT *root)
 {
   DBUG_ENTER("set_one_value");
   DBUG_PRINT("enter", ("opt: %p type: %u name '%s' value: '%s'",
-                       opt,
-                       opt->type, opt->name,
+                       opt, opt->type, opt->name,
                        (value->str ? value->str : "<DEFAULT>")));
   switch (opt->type)
   {
   case HA_OPTION_TYPE_SYSVAR:
-    DBUG_ASSERT(0); // HA_OPTION_TYPE_SYSVAR's are replaced in resolve_sysvars()
+    // HA_OPTION_TYPE_SYSVAR's are replaced in resolve_sysvars()
+    break; // to DBUG_ASSERT(0)
   case HA_OPTION_TYPE_ULL:
     {
       ulonglong *val= (ulonglong*)value_ptr(base, opt);
@@ -142,10 +140,9 @@ static bool set_one_value(ha_create_table_option *opt,
         DBUG_RETURN(0);
       }
 
-      my_option optp=
-        { opt->name, 1, 0, (uchar **)val, 0, 0, GET_ULL,
+      my_option optp= { opt->name, 1, 0, (uchar **)val, 0, 0, GET_ULL,
           REQUIRED_ARG, (longlong)opt->def_value, (longlong)opt->min_value,
-          opt->max_value, 0, (long) opt->block_size, 0};
+          opt->max_value, 0, (long) opt->block_size, 0 };
 
       ulonglong orig_val= strtoull(value->str, NULL, 10);
       my_bool unused;
@@ -172,31 +169,17 @@ static bool set_one_value(ha_create_table_option *opt,
     }
   case HA_OPTION_TYPE_ENUM:
     {
-      uint *val= (uint *)value_ptr(base, opt), num;
+      uint *val= (uint *)value_ptr(base, opt);
 
       *val= (uint) opt->def_value;
       if (!value->str)
         DBUG_RETURN(0);
 
-      const char *start= opt->values, *end;
-
-      num= 0;
-      while (*start)
+      uint num= value->find_in_list(opt->values);
+      if (num != UINT_MAX)
       {
-        for (end=start;
-             *end && *end != ',';
-             end++) /* no-op */;
-        if (!my_strnncoll(system_charset_info,
-                          (uchar*)start, end-start,
-                          (uchar*)value->str, value->length))
-        {
-          *val= num;
-          DBUG_RETURN(0);
-        }
-        if (*end)
-          end++;
-        start= end;
-        num++;
+        *val= num;
+        DBUG_RETURN(0);
       }
 
       DBUG_RETURN(report_wrong_value(thd, opt->name, value->str,
@@ -210,32 +193,11 @@ static bool set_one_value(ha_create_table_option *opt,
       if (!value->str)
         DBUG_RETURN(0);
 
-      if (!my_strnncoll(system_charset_info,
-                        (const uchar*)"NO", 2,
-                        (uchar *)value->str, value->length) ||
-          !my_strnncoll(system_charset_info,
-                        (const uchar*)"OFF", 3,
-                        (uchar *)value->str, value->length) ||
-          !my_strnncoll(system_charset_info,
-                        (const uchar*)"0", 1,
-                        (uchar *)value->str, value->length))
+      uint num= value->find_in_list(bools);
+      if (num != UINT_MAX)
       {
-        *val= FALSE;
-        DBUG_RETURN(FALSE);
-      }
-
-      if (!my_strnncoll(system_charset_info,
-                        (const uchar*)"YES", 3,
-                        (uchar *)value->str, value->length) ||
-          !my_strnncoll(system_charset_info,
-                        (const uchar*)"ON", 2,
-                        (uchar *)value->str, value->length) ||
-          !my_strnncoll(system_charset_info,
-                        (const uchar*)"1", 1,
-                        (uchar *)value->str, value->length))
-      {
-        *val= TRUE;
-        DBUG_RETURN(FALSE);
+        *val= num > 2;
+        DBUG_RETURN(0);
       }
 
       DBUG_RETURN(report_wrong_value(thd, opt->name, value->str,
@@ -250,6 +212,74 @@ static bool set_one_value(ha_create_table_option *opt,
 static const size_t ha_option_type_sizeof[]=
 { sizeof(ulonglong), sizeof(char *), sizeof(uint), sizeof(bool)};
 
+/**
+  Appends values of sysvar-based options if needed
+
+  @param thd              thread handler
+  @param option_list      list of options given by user
+  @param rules            list of option description by engine
+  @param root             MEM_ROOT where allocate memory
+
+  @retval TRUE  Error
+  @retval FALSE OK
+*/
+
+bool extend_option_list(THD* thd, st_plugin_int *plugin, bool create,
+                       engine_option_value **option_list,
+                       ha_create_table_option *rules)
+{
+  DBUG_ENTER("extend_option_list");
+  MEM_ROOT *root= thd->mem_root;
+  bool extended= false;
+
+  for (ha_create_table_option *opt= rules; rules && opt->name; opt++)
+  {
+    if (opt->var)
+    {
+      engine_option_value *found= NULL, *last;
+      for (engine_option_value *val= *option_list; val; val= val->next)
+      {
+        last= val;
+        if (val->name.streq(Lex_cstring(opt->name, opt->name_length)))
+          found= val; // find the last matching
+      }
+      if (found ? !found->value.str : create)
+      {
+        /* add the current value of the corresponding sysvar to the list */
+        sys_var *sysvar= find_plugin_sysvar(plugin, opt->var);
+        DBUG_ASSERT(sysvar);
+
+        if (!sysvar->session_is_default(thd))
+        {
+          StringBuffer<256> sbuf(system_charset_info);
+          String *str= sysvar->val_str(&sbuf, thd, OPT_SESSION, &null_clex_str);
+          DBUG_ASSERT(str);
+          engine_option_value::Name name(opt->name, opt->name_length);
+          engine_option_value::Value value;
+          value.str= strmake_root(root, str->ptr(), str->length());
+          value.length= str->length();
+          if (found)
+            found->value= value;
+          else
+          {
+            engine_option_value *val= new (root) engine_option_value(name,
+                                        value, opt->type != HA_OPTION_TYPE_ULL);
+            if (!extended)
+            {
+              if (*option_list)
+                thd->register_item_tree_change((Item**)&(last->next));
+              extended= true;
+            }
+            val->link(option_list, &last);
+          }
+        }
+      }
+    }
+  }
+  DBUG_RETURN(FALSE);
+}
+
+
 /**
   Creates option structure and parses list of options in it
 
@@ -264,7 +294,7 @@ static const size_t ha_option_type_sizeof[]=
   @retval FALSE OK
 */
 
-bool parse_option_list(THD* thd, handlerton *hton, void *option_struct_arg,
+bool parse_option_list(THD* thd, void *option_struct_arg,
                        engine_option_value **option_list,
                        ha_create_table_option *rules,
                        bool suppress_warning, MEM_ROOT *root)
@@ -273,6 +303,7 @@ bool parse_option_list(THD* thd, handlerton *hton, void *option_struct_arg,
   size_t option_struct_size= 0;
   engine_option_value *val, *last;
   void **option_struct= (void**)option_struct_arg;
+  engine_option_value::Value default_value;
   DBUG_ENTER("parse_option_list");
   DBUG_PRINT("enter",
              ("struct: %p list: %p rules: %p suppress_warning: %u root: %p",
@@ -294,9 +325,7 @@ bool parse_option_list(THD* thd, handlerton *hton, void *option_struct_arg,
     for (val= *option_list; val; val= val->next)
     {
       last= val;
-      if (my_strnncoll(system_charset_info,
-                       (uchar*)opt->name, opt->name_length,
-                       (uchar*)val->name.str, val->name.length))
+      if (!val->name.streq(Lex_cstring(opt->name, opt->name_length)))
         continue;
 
       /* skip duplicates (see engine_option_value constructor above) */
@@ -311,58 +340,8 @@ bool parse_option_list(THD* thd, handlerton *hton, void *option_struct_arg,
       break;
     }
     if (!seen || (opt->var && !last->value.str))
-    {
-      LEX_CSTRING default_val= null_clex_str;
-
-      /*
-        Okay, here's the logic for sysvar options:
-        1. When we parse CREATE TABLE and sysvar option was not explicitly
-           mentioned we add it to the list as if it was specified with the
-           *current* value of the underlying sysvar.
-        2. But only if the underlying sysvar value is different from the
-           sysvar's default.
-        3. If it's ALTER TABLE or CREATE_SEQUENCE and the sysvar option was
-           not explicitly mentioned - do nothing, do not add it to the list.
-        4. But if it was ALTER TABLE with sysvar option = DEFAULT, we
-           add it to the list (under the same condition #2).
-        5. If we're here parsing the option list from the .frm file
-           for a normal open_table() and the sysvar option was not there -
-           do not add it to the list (makes no sense anyway) and
-           use the *default* value of the underlying sysvar. Because
-           sysvar value can change, but it should not affect existing tables.
-        This is how it's implemented: the current sysvar value is added
-        to the list if suppress_warning is FALSE (meaning a table is created,
-        that is CREATE TABLE or ALTER TABLE) and it's actually a CREATE TABLE
-        command or it's an ALTER TABLE and the option was seen (=DEFAULT).
-
-        Note that if the option was set explicitly (not =DEFAULT) it wouldn't
-        have passes the if() condition above.
-      */
-      if (!suppress_warning && opt->var &&
-          (thd->lex->sql_command == SQLCOM_CREATE_TABLE || seen))
-      {
-        // take a value from the variable and add it to the list
-        sys_var *sysvar= find_hton_sysvar(hton, opt->var);
-        DBUG_ASSERT(sysvar);
-
-        if (!sysvar->session_is_default(thd))
-        {
-          char buf[256];
-          String sbuf(buf, sizeof(buf), system_charset_info), *str;
-          if ((str= sysvar->val_str(&sbuf, thd, OPT_SESSION, &null_clex_str)))
-          {
-            LEX_CSTRING name= { opt->name, opt->name_length };
-            default_val.str= strmake_root(root, str->ptr(), str->length());
-            default_val.length= str->length();
-            val= new (root) engine_option_value(name, default_val,
-                         opt->type != HA_OPTION_TYPE_ULL, option_list, &last);
-            val->parsed= true;
-          }
-        }
-      }
-      set_one_value(opt, thd, &default_val, *option_struct,
+      set_one_value(opt, thd, &default_value, *option_struct,
                     suppress_warning, root);
-    }
   }
 
   for (val= *option_list; val; val= val->next)
@@ -381,7 +360,7 @@ bool parse_option_list(THD* thd, handlerton *hton, void *option_struct_arg,
 
   This is done when an engine is loaded.
 */
-static bool resolve_sysvars(handlerton *hton, ha_create_table_option *rules)
+bool resolve_sysvar_table_options(ha_create_table_option *rules)
 {
   for (ha_create_table_option *opt= rules; rules && opt->name; opt++)
   {
@@ -417,11 +396,11 @@ static bool resolve_sysvars(handlerton *hton, ha_create_table_option *rules)
         str.length(0);
         for (const char **s= optp.typelib->type_names; *s; s++)
         {
-          if (str.append(*s) || str.append(','))
+          if (str.append(*s, strlen(*s)) || str.append(','))
             return 1;
         }
         DBUG_ASSERT(str.length());
-        opt->values= my_strndup(str.ptr(), str.length()-1, MYF(MY_WME));
+        opt->values= my_strndup(PSI_INSTRUMENT_ME, str.ptr(), str.length()-1, MYF(MY_WME));
         if (!opt->values)
           return 1;
         break;
@@ -434,13 +413,6 @@ static bool resolve_sysvars(handlerton *hton, ha_create_table_option *rules)
   return 0;
 }
 
-bool resolve_sysvar_table_options(handlerton *hton)
-{
-  return resolve_sysvars(hton, hton->table_options) ||
-         resolve_sysvars(hton, hton->field_options) ||
-         resolve_sysvars(hton, hton->index_options);
-}
-
 /*
   Restore HA_OPTION_TYPE_SYSVAR options back as they were
   before resolve_sysvars().
@@ -448,7 +420,7 @@ bool resolve_sysvar_table_options(handlerton *hton)
   This is done when the engine is unloaded, so that we could
   call resolve_sysvars() if the engine is installed again.
 */
-static void free_sysvars(handlerton *hton, ha_create_table_option *rules)
+void free_sysvar_table_options(ha_create_table_option *rules)
 {
   for (ha_create_table_option *opt= rules; rules && opt->name; opt++)
   {
@@ -465,14 +437,6 @@ static void free_sysvars(handlerton *hton, ha_create_table_option *rules)
   }
 }
 
-void free_sysvar_table_options(handlerton *hton)
-{
-  free_sysvars(hton, hton->table_options);
-  free_sysvars(hton, hton->field_options);
-  free_sysvars(hton, hton->index_options);
-}
-
-
 /**
   Parses all table/fields/keys options
 
@@ -489,13 +453,13 @@ bool parse_engine_table_options(THD *thd, handlerton *ht, TABLE_SHARE *share)
   MEM_ROOT *root= &share->mem_root;
   DBUG_ENTER("parse_engine_table_options");
 
-  if (parse_option_list(thd, ht, &share->option_struct, & share->option_list,
+  if (parse_option_list(thd, &share->option_struct, & share->option_list,
                         ht->table_options, TRUE, root))
     DBUG_RETURN(TRUE);
 
   for (Field **field= share->field; *field; field++)
   {
-    if (parse_option_list(thd, ht, &(*field)->option_struct,
+    if (parse_option_list(thd, &(*field)->option_struct,
                           & (*field)->option_list,
                           ht->field_options, TRUE, root))
       DBUG_RETURN(TRUE);
@@ -503,7 +467,7 @@ bool parse_engine_table_options(THD *thd, handlerton *ht, TABLE_SHARE *share)
 
   for (uint index= 0; index < share->keys; index ++)
   {
-    if (parse_option_list(thd, ht, &share->key_info[index].option_struct,
+    if (parse_option_list(thd, &share->key_info[index].option_struct,
                           & share->key_info[index].option_list,
                           ht->index_options, TRUE, root))
       DBUG_RETURN(TRUE);
@@ -512,6 +476,61 @@ bool parse_engine_table_options(THD *thd, handlerton *ht, TABLE_SHARE *share)
   DBUG_RETURN(FALSE);
 }
 
+#ifdef WITH_PARTITION_STORAGE_ENGINE
+/**
+  Parses engine-defined partition options
+
+  @param [in] thd   thread handler
+  @parem [in] table table with part_info
+
+  @retval TRUE  Error
+  @retval FALSE OK
+
+  In the case of ALTER TABLE statements, table->part_info is set up
+  by mysql_unpack_partition(). So, one should not call the present
+  function before the call of mysql_unpack_partition().
+*/
+bool parse_engine_part_options(THD *thd, TABLE *table)
+{
+  MEM_ROOT *root= &table->mem_root;
+  TABLE_SHARE *share= table->s;
+  partition_info *part_info= table->part_info;
+  engine_option_value *tmp_option_list;
+  handlerton *ht;
+  DBUG_ENTER("parse_engine_part_options");
+
+  if (!part_info)
+    DBUG_RETURN(FALSE);
+
+  List_iterator<partition_element> it(part_info->partitions);
+  while (partition_element *part_elem= it++)
+  {
+    if (merge_engine_options(share->option_list, part_elem->option_list,
+                             &tmp_option_list, root))
+      DBUG_RETURN(TRUE);
+
+    if (!part_info->is_sub_partitioned())
+    {
+      ht= part_elem->engine_type;
+      if (parse_option_list(thd, &part_elem->option_struct,
+                            &tmp_option_list, ht->table_options, TRUE, root))
+        DBUG_RETURN(TRUE);
+    }
+    else
+    {
+      List_iterator<partition_element> sub_it(part_elem->subpartitions);
+      while (partition_element *sub_part_elem= sub_it++)
+      {
+        ht= sub_part_elem->engine_type;
+        if (parse_option_list(thd, &sub_part_elem->option_struct,
+                              &tmp_option_list, ht->table_options, TRUE, root))
+          DBUG_RETURN(TRUE);
+      }
+    }
+  }
+  DBUG_RETURN(FALSE);
+}
+#endif
 
 bool engine_options_differ(void *old_struct, void *new_struct,
                            ha_create_table_option *rules)
@@ -709,10 +728,13 @@ uchar *engine_option_value::frm_read(const uchar *buff, const uchar *buff_end,
     return NULL;
   buff+= value.length;
 
-  engine_option_value *ptr=new (root)
-    engine_option_value(name, value, len & FRM_QUOTED_VALUE, start, end);
+  engine_option_value *ptr=
+      new (root) engine_option_value(engine_option_value::Name(name),
+                                     engine_option_value::Value(value),
+                                     len & FRM_QUOTED_VALUE);
   if (!ptr)
     return NULL;
+  ptr->link(start, end);
 
   return (uchar *)buff;
 }
@@ -759,7 +781,7 @@ bool engine_table_options_frm_read(const uchar *buff, size_t length,
     buff++;
   }
 
-  for (count=0; count < share->keys; count++)
+  for (count=0; count < share->total_keys; count++)
   {
     while (buff < buff_end && *buff)
     {
@@ -781,23 +803,40 @@ bool engine_table_options_frm_read(const uchar *buff, size_t length,
 
 /**
   Merges two lists of engine_option_value's with duplicate removal.
+
+  @param [in] source  option list
+  @param [in] changes option list whose options overwrite source's
+  @param [out] out    new option list created by merging given two
+  @param [in] root    MEM_ROOT for allocating memory
+
+  @retval TRUE  Error
+  @retval FALSE OK
 */
-
-engine_option_value *merge_engine_table_options(engine_option_value *first,
-                                                engine_option_value *second,
-                                                MEM_ROOT *root)
+bool merge_engine_options(engine_option_value *source,
+                          engine_option_value *changes,
+                          engine_option_value **out, MEM_ROOT *root)
 {
-  engine_option_value *UNINIT_VAR(end), *opt;
-  DBUG_ENTER("merge_engine_table_options");
+  engine_option_value *UNINIT_VAR(end), *opt, *opt_copy;
+  *out= 0;
+  DBUG_ENTER("merge_engine_options");
 
-  /* Create copy of first list */
-  for (opt= first, first= 0; opt; opt= opt->next)
-    new (root) engine_option_value(opt, &first, &end);
+  /* Create copy of source list */
+  for (opt= source; opt; opt= opt->next)
+  {
+    opt_copy= new (root) engine_option_value(opt);
+    if (!opt_copy)
+      DBUG_RETURN(TRUE);
+    opt_copy->link(out, &end);
+  }
 
-  for (opt= second; opt; opt= opt->next)
-    new (root) engine_option_value(opt->name, opt->value, opt->quoted_value,
-                                   &first, &end);
-  DBUG_RETURN(first);
+  for (opt= changes; opt; opt= opt->next)
+  {
+    opt_copy= new (root) engine_option_value(opt);
+    if (!opt_copy)
+      DBUG_RETURN(TRUE);
+    opt_copy->link(out, &end);
+  }
+  DBUG_RETURN(FALSE);
 }
 
 bool is_engine_option_known(engine_option_value *opt,
@@ -808,11 +847,8 @@ bool is_engine_option_known(engine_option_value *opt,
 
   for (; rules->name; rules++)
   {
-      if (!my_strnncoll(system_charset_info,
-                        (uchar*)rules->name, rules->name_length,
-                        (uchar*)opt->name.str, opt->name.length))
+      if (opt->name.streq(Lex_cstring(rules->name, rules->name_length)))
         return true;
   }
   return false;
 }
-
