@@ -29,37 +29,38 @@ Window_spec::check_window_names(List_iterator_fast<Window_spec> &it)
 {
   if (window_names_are_checked)
     return false;
-  const char *name= this->name();
-  const char *ref_name= window_reference();
+  const Lex_ident_window name= this->name();
+  const Lex_ident_window ref_name= window_reference();
   it.rewind();
   Window_spec *win_spec;
   while((win_spec= it++) && win_spec != this)
   {
-    const char *win_spec_name= win_spec->name();
-    if (!win_spec_name)
+    const Lex_ident_window win_spec_name= win_spec->name();
+    if (!win_spec_name.str)
       break;
-    if (name && my_strcasecmp(system_charset_info, name, win_spec_name) == 0)
+    if (name.str && name.streq(win_spec_name))
     {
-      my_error(ER_DUP_WINDOW_NAME, MYF(0), name);
+      my_error(ER_DUP_WINDOW_NAME, MYF(0), name.str);
       return true;
     }
-    if (ref_name &&
-        my_strcasecmp(system_charset_info, ref_name, win_spec_name) == 0)
+    if (ref_name.str && win_spec_name.streq(ref_name))
     {
       if (partition_list->elements)
       {
         my_error(ER_PARTITION_LIST_IN_REFERENCING_WINDOW_SPEC, MYF(0),
-                 ref_name);
+                 ref_name.str);
         return true;
       }
       if (win_spec->order_list->elements && order_list->elements)
       {
-        my_error(ER_ORDER_LIST_IN_REFERENCING_WINDOW_SPEC, MYF(0), ref_name);
+        my_error(ER_ORDER_LIST_IN_REFERENCING_WINDOW_SPEC, MYF(0),
+                 ref_name.str);
         return true;
       } 
       if (win_spec->window_frame)
       {
-        my_error(ER_WINDOW_FRAME_IN_REFERENCED_WINDOW_SPEC, MYF(0), ref_name);
+        my_error(ER_WINDOW_FRAME_IN_REFERENCED_WINDOW_SPEC, MYF(0),
+                 ref_name.str);
         return true;
       }
       referenced_win_spec= win_spec;
@@ -69,9 +70,9 @@ Window_spec::check_window_names(List_iterator_fast<Window_spec> &it)
         order_list= win_spec->order_list;
     }
   }
-  if (ref_name && !referenced_win_spec)
+  if (ref_name.str && !referenced_win_spec)
   {
-    my_error(ER_WRONG_WINDOW_SPEC_NAME, MYF(0), ref_name);
+    my_error(ER_WRONG_WINDOW_SPEC_NAME, MYF(0), ref_name.str);
     return true;
   }
   window_names_are_checked= true;
@@ -221,7 +222,7 @@ setup_windows(THD *thd, Ref_ptr_array ref_pointer_array, TABLE_LIST *tables,
     uint elems= win_specs.elements;
     while ((win_spec= it++) && i++ < elems)
     {
-      if (win_spec->name() == NULL)
+      if (win_spec->name().str == NULL)
       {
         it.remove();
         win_specs.push_back(win_spec);
@@ -432,6 +433,16 @@ ORDER *st_select_lex::find_common_window_func_partition_fields(THD *thd)
 #define CMP_GT_C       1    // Greater than and compatible
 #define CMP_GT         2    // Greater then
 
+
+/*
+  This function is used for sorting ORDER/PARTITION BY clauses of window
+  functions and so must implement an order relation on ORDER BY clauses"
+
+  It is called by a sorting function.
+  The function return's CMP_EQ (=0) if the values are identical.
+  If not equal, it returns a stable value < or > than 0.
+*/
+
 static
 int compare_order_elements(ORDER *ord1, int weight1,
                            ORDER *ord2, int weight2)
@@ -640,7 +651,7 @@ int compare_window_funcs_by_window_specs(Item_window_func *win_func1,
       Partition lists contain the same elements. 
       Let's use only one of the lists.
     */
-    if (!win_spec1->name() && win_spec2->name())
+    if (!win_spec1->name().str && win_spec2->name().str)
     {
       win_spec1->save_partition_list= win_spec1->partition_list;
       win_spec1->partition_list= win_spec2->partition_list;
@@ -663,14 +674,14 @@ int compare_window_funcs_by_window_specs(Item_window_func *win_func1,
        Order lists contain the same elements.
        Let's use only one of the lists.
     */
-    if (!win_spec1->name() && win_spec2->name())
+    if (!win_spec1->name().str && win_spec2->name().str)
     {
-      win_spec1->save_order_list= win_spec2->order_list;
+      win_spec1->save_order_list= win_spec1->order_list;
       win_spec1->order_list= win_spec2->order_list;
     }
     else
     {
-      win_spec1->save_order_list= win_spec2->order_list;
+      win_spec2->save_order_list= win_spec2->order_list;
       win_spec2->order_list= win_spec1->order_list;
     }
 
@@ -681,7 +692,7 @@ int compare_window_funcs_by_window_specs(Item_window_func *win_func1,
       return cmp;
 
     /* Window frames are equal. Let's use only one of them. */
-    if (!win_spec1->name() && win_spec2->name())
+    if (!win_spec1->name().str && win_spec2->name().str)
       win_spec1->window_frame= win_spec2->window_frame;
     else
       win_spec2->window_frame= win_spec1->window_frame;
@@ -703,10 +714,6 @@ int compare_window_funcs_by_window_specs(Item_window_func *win_func1,
 }
 
 
-#define  SORTORDER_CHANGE_FLAG    1
-#define  PARTITION_CHANGE_FLAG    2
-#define  FRAME_CHANGE_FLAG        4
-
 typedef int (*Item_window_func_cmp)(Item_window_func *f1,
                                     Item_window_func *f2,
                                     void *arg); 
@@ -736,15 +743,15 @@ void order_window_funcs_by_window_specs(List<Item_window_func> *win_func_list)
 
   List_iterator_fast<Item_window_func> it(*win_func_list);
   Item_window_func *prev= it++;
-  prev->marker= SORTORDER_CHANGE_FLAG |
-                PARTITION_CHANGE_FLAG |
-                FRAME_CHANGE_FLAG;
+  prev->marker= (MARKER_SORTORDER_CHANGE |
+                 MARKER_PARTITION_CHANGE |
+                 MARKER_FRAME_CHANGE);
   Item_window_func *curr;
   while ((curr= it++))
   {
     Window_spec *win_spec_prev= prev->window_spec;
     Window_spec *win_spec_curr= curr->window_spec;
-    curr->marker= 0;
+    curr->marker= MARKER_UNUSED;
     if (!(win_spec_prev->partition_list == win_spec_curr->partition_list &&
           win_spec_prev->order_list == win_spec_curr->order_list))
     {
@@ -758,17 +765,17 @@ void order_window_funcs_by_window_specs(List<Item_window_func> *win_func_list)
         cmp= compare_window_spec_joined_lists(win_spec_prev, win_spec_curr);
       if (!(CMP_LT_C <= cmp && cmp <= CMP_GT_C))
       {
-        curr->marker= SORTORDER_CHANGE_FLAG |
-                      PARTITION_CHANGE_FLAG |
-                      FRAME_CHANGE_FLAG;
+        curr->marker= (MARKER_SORTORDER_CHANGE |
+                       MARKER_PARTITION_CHANGE |
+                       MARKER_FRAME_CHANGE);
       }
       else if (win_spec_prev->partition_list != win_spec_curr->partition_list)
       {
-        curr->marker|= PARTITION_CHANGE_FLAG | FRAME_CHANGE_FLAG;
+        curr->marker|= MARKER_PARTITION_CHANGE | MARKER_FRAME_CHANGE;
       }
     }
     else if (win_spec_prev->window_frame != win_spec_curr->window_frame)
-      curr->marker|= FRAME_CHANGE_FLAG;
+      curr->marker|= MARKER_FRAME_CHANGE;
 
     prev= curr;
   }
@@ -2494,13 +2501,25 @@ Frame_cursor *get_frame_cursor(THD *thd, Window_spec *spec, bool is_top_bound)
     }
     else
     {
+      /*
+        compare_window_funcs_by_window_specs() will try to get the 
+        Window Specs to reuse the ORDER BY lists.
+        RANGE-type window frame expects a single ORDER BY element,
+        and if the list from a different window spec having more than 1 
+        ORDER BY element is used, then an ASSERT is raised.
+        
+        So, use the original ORDER BY list when constructing 
+        RANGE-type frames.
+      */
+      SQL_I_List<ORDER> *order_list=
+          spec->save_order_list ? spec->save_order_list : spec->order_list;
       if (is_top_bound)
         return new Frame_range_n_top(
-            thd, spec->partition_list, spec->order_list,
+            thd, spec->partition_list, order_list,
             is_preceding, bound->offset);
 
       return new Frame_range_n_bottom(thd,
-          spec->partition_list, spec->order_list,
+          spec->partition_list, order_list,
           is_preceding, bound->offset);
     }
   }
@@ -2562,7 +2581,7 @@ bool add_special_frame_cursors(THD *thd, Cursor_manager *cursor_manager,
       cursor_manager->add_cursor(bottom_bound);
       cursor_manager->add_cursor(top_bound);
       cursor_manager->add_cursor(current_row_pos);
-      DBUG_ASSERT(item_sum->fixed);
+      DBUG_ASSERT(item_sum->fixed());
       bool negative_offset= item_sum->sum_func() == Item_sum::LAG_FUNC;
       fc= new Frame_positional_cursor(*current_row_pos,
                                       *top_bound, *bottom_bound,
@@ -2578,7 +2597,7 @@ bool add_special_frame_cursors(THD *thd, Cursor_manager *cursor_manager,
       Frame_cursor *top_bound= get_frame_cursor(thd, spec, true);
       cursor_manager->add_cursor(bottom_bound);
       cursor_manager->add_cursor(top_bound);
-      DBUG_ASSERT(item_sum->fixed);
+      DBUG_ASSERT(item_sum->fixed());
       Item *offset_item= new (thd->mem_root) Item_int(thd, 0);
       offset_item->fix_fields(thd, &offset_item);
       fc= new Frame_positional_cursor(*top_bound,
@@ -2594,7 +2613,7 @@ bool add_special_frame_cursors(THD *thd, Cursor_manager *cursor_manager,
       Frame_cursor *top_bound= get_frame_cursor(thd, spec, true);
       cursor_manager->add_cursor(bottom_bound);
       cursor_manager->add_cursor(top_bound);
-      DBUG_ASSERT(item_sum->fixed);
+      DBUG_ASSERT(item_sum->fixed());
       Item *offset_item= new (thd->mem_root) Item_int(thd, 0);
       offset_item->fix_fields(thd, &offset_item);
       fc= new Frame_positional_cursor(*bottom_bound,
@@ -2610,7 +2629,7 @@ bool add_special_frame_cursors(THD *thd, Cursor_manager *cursor_manager,
       Frame_cursor *top_bound= get_frame_cursor(thd, spec, true);
       cursor_manager->add_cursor(bottom_bound);
       cursor_manager->add_cursor(top_bound);
-      DBUG_ASSERT(item_sum->fixed);
+      DBUG_ASSERT(item_sum->fixed());
       Item *int_item= new (thd->mem_root) Item_int(thd, 1);
       Item *offset_func= new (thd->mem_root)
                               Item_func_minus(thd, item_sum->get_arg(1),
@@ -2807,7 +2826,7 @@ bool save_window_function_values(List<Item_window_func>& window_functions,
   Item *func;
   for (; (func = *func_ptr) ; func_ptr++)
   {
-    if (func->with_window_func && func->type() != Item::WINDOW_FUNC_ITEM)
+    if (func->with_window_func() && func->type() != Item::WINDOW_FUNC_ITEM)
       func->save_in_result_field(true);
   }
 
@@ -2863,6 +2882,7 @@ bool compute_window_func(THD *thd,
 {
   List_iterator_fast<Item_window_func> iter_win_funcs(window_functions);
   List_iterator_fast<Cursor_manager> iter_cursor_managers(cursor_managers);
+  bool ret= false;
   uint err;
 
   READ_RECORD info;
@@ -2925,18 +2945,27 @@ bool compute_window_func(THD *thd,
       /* Check if we found any error in the window function while adding values
          through cursors. */
       if (unlikely(thd->is_error() || thd->is_killed()))
+      {
+        ret= true;
         break;
-
+      }
 
       /* Return to current row after notifying cursors for each window
          function. */
-      tbl->file->ha_rnd_pos(tbl->record[0], rowid_buf);
+      if (tbl->file->ha_rnd_pos(tbl->record[0], rowid_buf))
+      {
+        ret= true;
+        break;
+      }
     }
 
     /* We now have computed values for each window function. They can now
        be saved in the current row. */
-    save_window_function_values(window_functions, tbl, rowid_buf);
-
+    if (save_window_function_values(window_functions, tbl, rowid_buf))
+    {
+      ret= true;
+      break;
+    }
     rownum++;
   }
 
@@ -2944,7 +2973,7 @@ bool compute_window_func(THD *thd,
   partition_trackers.delete_elements();
   end_read_record(&info);
 
-  return false;
+  return ret;
 }
 
 /* Make a list that is a concation of two lists of ORDER elements */
@@ -3037,7 +3066,8 @@ bool Window_func_runner::exec(THD *thd, TABLE *tbl, SORT_INFO *filesort_result)
     win_func->set_phase_to_computation();
     // TODO(cvicentiu) Setting the aggregator should probably be done during
     // setup of Window_funcs_sort.
-    win_func->window_func()->set_aggregator(Aggregator::SIMPLE_AGGREGATOR);
+    win_func->window_func()->set_aggregator(thd,
+                                            Aggregator::SIMPLE_AGGREGATOR);
   }
   it.rewind();
 
@@ -3111,11 +3141,11 @@ bool Window_funcs_sort::setup(THD *thd, SQL_SELECT *sel,
       return true;
     it++;
     win_func= it.peek();
-  } while (win_func && !(win_func->marker & SORTORDER_CHANGE_FLAG));
+  } while (win_func && !(win_func->marker & MARKER_SORTORDER_CHANGE));
 
   /*
     The sort criteria must be taken from the last win_func in the group of
-    adjacent win_funcs that do not have SORTORDER_CHANGE_FLAG. This is
+    adjacent win_funcs that do not have MARKER_SORTORDER_CHANGE. This is
     because the sort order must be the most specific sorting criteria defined
     within the window function group. This ensures that we sort the table
     in a way that the result is valid for all window functions belonging to
@@ -3140,8 +3170,10 @@ bool Window_funcs_sort::setup(THD *thd, SQL_SELECT *sel,
      */
     ORDER *order= (ORDER *)alloc_root(thd->mem_root, sizeof(ORDER));
     memset(order, 0, sizeof(*order));
-    Item *item= new (thd->mem_root) Item_temptable_field(thd,
-                                                    join_tab->table->field[0]);
+    Item_field *item=
+        new (thd->mem_root) Item_field(thd, join_tab->table->field[0]);
+    if (item)
+      item->set_refers_to_temp_table();
     order->item= (Item **)alloc_root(thd->mem_root, 2 * sizeof(Item *));
     order->item[1]= NULL;
     order->item[0]= item;
