@@ -2,7 +2,7 @@
 
 Copyright (c) 2010, 2015, Oracle and/or its affiliates. All Rights Reserved.
 Copyright (c) 2012, Facebook Inc.
-Copyright (c) 2013, 2022, MariaDB Corporation.
+Copyright (c) 2013, 2023, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
@@ -36,7 +36,7 @@ Created 12/15/2009	Jimmy Yang
 #define __STDC_LIMIT_MACROS
 #endif /* __STDC_LIMIT_MACROS */
 
-#include <stdint.h>
+#include <cstdint>
 #include "my_atomic.h"
 #include "my_atomic_wrapper.h"
 
@@ -136,8 +136,6 @@ enum monitor_id_t {
 	/* Start of Metadata counter */
 	MONITOR_MODULE_METADATA,
 	MONITOR_TABLE_OPEN,
-	MONITOR_TABLE_CLOSE,
-	MONITOR_TABLE_REFERENCE,
 
 	/* Lock manager related counters */
 	MONITOR_MODULE_LOCK,
@@ -177,8 +175,6 @@ enum monitor_id_t {
 	MONITOR_OVLD_PAGE_CREATED,
 	MONITOR_OVLD_PAGES_WRITTEN,
 	MONITOR_OVLD_PAGES_READ,
-	MONITOR_OVLD_INDEX_SEC_REC_CLUSTER_READS,
-	MONITOR_OVLD_INDEX_SEC_REC_CLUSTER_READS_AVOIDED,
 	MONITOR_OVLD_BYTE_READ,
 	MONITOR_OVLD_BYTE_WRITTEN,
 	MONITOR_FLUSH_BATCH_SCANNED,
@@ -198,7 +194,6 @@ enum monitor_id_t {
 	MONITOR_FLUSH_ADAPTIVE_AVG_PASS,
 
 	MONITOR_LRU_GET_FREE_LOOPS,
-	MONITOR_LRU_GET_FREE_WAITS,
 
 	MONITOR_FLUSH_AVG_PAGE_RATE,
 	MONITOR_FLUSH_LSN_AVG_RATE,
@@ -218,12 +213,7 @@ enum monitor_id_t {
 	MONITOR_LRU_BATCH_SCANNED_NUM_CALL,
 	MONITOR_LRU_BATCH_SCANNED_PER_CALL,
 	MONITOR_LRU_BATCH_FLUSH_TOTAL_PAGE,
-	MONITOR_LRU_BATCH_FLUSH_COUNT,
-	MONITOR_LRU_BATCH_FLUSH_PAGES,
 	MONITOR_LRU_BATCH_EVICT_TOTAL_PAGE,
-	MONITOR_LRU_BATCH_EVICT_COUNT,
-	MONITOR_LRU_BATCH_EVICT_PAGES,
-	MONITOR_LRU_SINGLE_FLUSH_FAILURE_COUNT,
 	MONITOR_LRU_GET_FREE_SEARCH,
 	MONITOR_LRU_SEARCH_SCANNED,
 	MONITOR_LRU_SEARCH_SCANNED_NUM_CALL,
@@ -236,12 +226,8 @@ enum monitor_id_t {
 	MONITOR_MODULE_BUF_PAGE,
 	MONITOR_INDEX_LEAF_PAGE_READ,
 	MONITOR_INDEX_NON_LEAF_PAGE_READ,
-	MONITOR_INDEX_IBUF_LEAF_PAGE_READ,
-	MONITOR_INDEX_IBUF_NON_LEAF_PAGE_READ,
 	MONITOR_UNDO_LOG_PAGE_READ,
 	MONITOR_INODE_PAGE_READ,
-	MONITOR_IBUF_FREELIST_PAGE_READ,
-	MONITOR_IBUF_BITMAP_PAGE_READ,
 	MONITOR_SYSTEM_PAGE_READ,
 	MONITOR_TRX_SYSTEM_PAGE_READ,
 	MONITOR_FSP_HDR_PAGE_READ,
@@ -252,12 +238,8 @@ enum monitor_id_t {
 	MONITOR_OTHER_PAGE_READ,
 	MONITOR_INDEX_LEAF_PAGE_WRITTEN,
 	MONITOR_INDEX_NON_LEAF_PAGE_WRITTEN,
-	MONITOR_INDEX_IBUF_LEAF_PAGE_WRITTEN,
-	MONITOR_INDEX_IBUF_NON_LEAF_PAGE_WRITTEN,
 	MONITOR_UNDO_LOG_PAGE_WRITTEN,
 	MONITOR_INODE_PAGE_WRITTEN,
-	MONITOR_IBUF_FREELIST_PAGE_WRITTEN,
-	MONITOR_IBUF_BITMAP_PAGE_WRITTEN,
 	MONITOR_SYSTEM_PAGE_WRITTEN,
 	MONITOR_TRX_SYSTEM_PAGE_WRITTEN,
 	MONITOR_FSP_HDR_PAGE_WRITTEN,
@@ -275,9 +257,6 @@ enum monitor_id_t {
 	MONITOR_OS_PENDING_READS,
 	MONITOR_OS_PENDING_WRITES,
 	MONITOR_OVLD_OS_LOG_WRITTEN,
-	MONITOR_OVLD_OS_LOG_FSYNC,
-	MONITOR_OVLD_OS_LOG_PENDING_FSYNC,
-	MONITOR_OVLD_OS_LOG_PENDING_WRITES,
 
 	/* Transaction related counters */
 	MONITOR_MODULE_TRX,
@@ -287,7 +266,6 @@ enum monitor_id_t {
 	MONITOR_TRX_COMMIT_UNDO,
 	MONITOR_TRX_ROLLBACK,
 	MONITOR_TRX_ROLLBACK_SAVEPOINT,
-	MONITOR_TRX_ACTIVE,
 	MONITOR_RSEG_HISTORY_LEN,
 	MONITOR_NUM_UNDO_SLOT_USED,
 	MONITOR_NUM_UNDO_SLOT_CACHED,
@@ -305,20 +283,16 @@ enum monitor_id_t {
 
 	/* Recovery related counters */
 	MONITOR_MODULE_RECOVERY,
-	MONITOR_NUM_CHECKPOINT,
+	MONITOR_OVLD_CHECKPOINTS,
 	MONITOR_OVLD_LSN_FLUSHDISK,
 	MONITOR_OVLD_LSN_CHECKPOINT,
 	MONITOR_OVLD_LSN_CURRENT,
 	MONITOR_LSN_CHECKPOINT_AGE,
 	MONITOR_OVLD_BUF_OLDEST_LSN,
 	MONITOR_OVLD_MAX_AGE_ASYNC,
-	MONITOR_PENDING_LOG_FLUSH,
-	MONITOR_PENDING_CHECKPOINT_WRITE,
-	MONITOR_LOG_IO,
 	MONITOR_OVLD_LOG_WAITS,
 	MONITOR_OVLD_LOG_WRITE_REQUEST,
 	MONITOR_OVLD_LOG_WRITES,
-	MONITOR_OVLD_LOG_PADDED,
 
 	/* Page Manager related counters */
 	MONITOR_MODULE_PAGE,
@@ -350,9 +324,7 @@ enum monitor_id_t {
 	/* Adaptive Hash Index related counters */
 	MONITOR_MODULE_ADAPTIVE_HASH,
 	MONITOR_OVLD_ADAPTIVE_HASH_SEARCH,
-#endif /* BTR_CUR_HASH_ADAPT */
 	MONITOR_OVLD_ADAPTIVE_HASH_SEARCH_BTREE,
-#ifdef BTR_CUR_HASH_ADAPT
 	MONITOR_ADAPTIVE_HASH_PAGE_ADDED,
 	MONITOR_ADAPTIVE_HASH_PAGE_REMOVED,
 	MONITOR_ADAPTIVE_HASH_ROW_ADDED,
@@ -365,24 +337,12 @@ enum monitor_id_t {
 	MONITOR_MODULE_FIL_SYSTEM,
 	MONITOR_OVLD_N_FILE_OPENED,
 
-	/* InnoDB Change Buffer related counters */
-	MONITOR_MODULE_IBUF_SYSTEM,
-	MONITOR_OVLD_IBUF_MERGE_INSERT,
-	MONITOR_OVLD_IBUF_MERGE_DELETE,
-	MONITOR_OVLD_IBUF_MERGE_PURGE,
-	MONITOR_OVLD_IBUF_MERGE_DISCARD_INSERT,
-	MONITOR_OVLD_IBUF_MERGE_DISCARD_DELETE,
-	MONITOR_OVLD_IBUF_MERGE_DISCARD_PURGE,
-	MONITOR_OVLD_IBUF_MERGES,
-	MONITOR_OVLD_IBUF_SIZE,
-
 	/* Counters for server operations */
 	MONITOR_MODULE_SERVER,
 	MONITOR_MASTER_THREAD_SLEEP,
 	MONITOR_OVLD_SERVER_ACTIVITY,
 	MONITOR_MASTER_ACTIVE_LOOPS,
 	MONITOR_MASTER_IDLE_LOOPS,
-	MONITOR_SRV_BACKGROUND_DROP_TABLE_MICROSECOND,
 	MONITOR_SRV_LOG_FLUSH_MICROSECOND,
 	MONITOR_SRV_DICT_LRU_MICROSECOND,
 	MONITOR_SRV_DICT_LRU_EVICT_COUNT_ACTIVE,
@@ -390,31 +350,10 @@ enum monitor_id_t {
 	MONITOR_OVLD_SRV_DBLWR_WRITES,
 	MONITOR_OVLD_SRV_DBLWR_PAGES_WRITTEN,
 	MONITOR_OVLD_SRV_PAGE_SIZE,
-	MONITOR_OVLD_RWLOCK_S_SPIN_WAITS,
-	MONITOR_OVLD_RWLOCK_X_SPIN_WAITS,
-	MONITOR_OVLD_RWLOCK_SX_SPIN_WAITS,
-	MONITOR_OVLD_RWLOCK_S_SPIN_ROUNDS,
-	MONITOR_OVLD_RWLOCK_X_SPIN_ROUNDS,
-	MONITOR_OVLD_RWLOCK_SX_SPIN_ROUNDS,
-	MONITOR_OVLD_RWLOCK_S_OS_WAITS,
-	MONITOR_OVLD_RWLOCK_X_OS_WAITS,
-	MONITOR_OVLD_RWLOCK_SX_OS_WAITS,
-
-	/* Data DML related counters */
-	MONITOR_MODULE_DML_STATS,
-	MONITOR_OLVD_ROW_READ,
-	MONITOR_OLVD_ROW_INSERTED,
-	MONITOR_OLVD_ROW_DELETED,
-	MONITOR_OLVD_ROW_UPDTATED,
-	MONITOR_OLVD_SYSTEM_ROW_READ,
-	MONITOR_OLVD_SYSTEM_ROW_INSERTED,
-	MONITOR_OLVD_SYSTEM_ROW_DELETED,
-	MONITOR_OLVD_SYSTEM_ROW_UPDATED,
 
 	/* Data DDL related counters */
 	MONITOR_MODULE_DDL_STATS,
 	MONITOR_BACKGROUND_DROP_INDEX,
-	MONITOR_BACKGROUND_DROP_TABLE,
 	MONITOR_ONLINE_CREATE_INDEX,
 	MONITOR_PENDING_ALTER_TABLE,
 	MONITOR_ALTER_TABLE_SORT_FILES,
@@ -426,10 +365,6 @@ enum monitor_id_t {
 	MONITOR_ICP_OUT_OF_RANGE,
 	MONITOR_ICP_MATCH,
 
-	/* Mutex/RW-Lock related counters */
-	MONITOR_MODULE_LATCHES,
-	MONITOR_LATCHES,
-
 	/* This is used only for control system to turn
 	on/off and reset all monitor counters */
 	MONITOR_ALL_COUNTER,
@@ -872,13 +807,9 @@ void
 srv_mon_reset(
 /*==========*/
 	monitor_id_t	monitor);	/*!< in: monitor id*/
-/*************************************************************//**
-This function resets all values of a monitor counter */
-UNIV_INLINE
-void
-srv_mon_reset_all(
-/*==============*/
-	monitor_id_t	monitor);	/*!< in: monitor id*/
+
+/** This function resets all values of a monitor counter */
+void srv_mon_reset_all(monitor_id_t monitor) noexcept;
 /*************************************************************//**
 Turn on monitor counters that are marked as default ON. */
 void
