@@ -15,6 +15,7 @@
 
 /* Describe, check and repair of MARIA tables */
 
+#define VER "1.3"
 #include "ma_fulltext.h"
 #include <myisamchk.h>
 #include <my_bit.h>
@@ -25,6 +26,7 @@
 /* Remove next line if you want aria_chk to produce a stack trace */
 #undef HAVE_BACKTRACE
 #include <my_stacktrace.h>
+#include <welcome_copyright_notice.h>
 
 static uint decode_bits;
 static char **default_argv;
@@ -79,7 +81,6 @@ static char default_open_errmsg[]=  "%d when opening Aria table '%s'";
 static char default_close_errmsg[]= "%d when closing Aria table '%s'";
 
 static void get_options(int *argc,char * * *argv);
-static void print_version(void);
 static void usage(void);
 static int maria_chk(HA_CHECK *param, char *filename);
 static void descript(HA_CHECK *param, register MARIA_HA *info, char *name);
@@ -145,7 +146,8 @@ int main(int argc, char **argv)
   {
     if ((ma_control_file_open(FALSE, opt_require_control_file ||
                               !(check_param.testflag & T_SILENT),
-                              TRUE)))
+                              TRUE,
+                              control_file_open_flags)))
     {
       if (opt_require_control_file ||
           (opt_transaction_logging && (check_param.testflag & T_REP_ANY)))
@@ -235,9 +237,6 @@ end:
   }
   maria_end();
   my_exit(error);
-#ifndef _lint
-  return 0;				/* No compiler warning */
-#endif
 } /* main */
 
 enum options_mc {
@@ -250,7 +249,7 @@ enum options_mc {
   OPT_MAX_RECORD_LENGTH, OPT_AUTO_CLOSE, OPT_STATS_METHOD, OPT_TRANSACTION_LOG,
   OPT_ZEROFILL_KEEP_LSN,
   OPT_REQUIRE_CONTROL_FILE, OPT_IGNORE_CONTROL_FILE,
-  OPT_LOG_DIR, OPT_WARNING_FOR_WRONG_TRANSID
+  OPT_LOG_DIR, OPT_WARNING_FOR_WRONG_TRANSID,OPT_ACTIVE_KEYS
 };
 
 static struct my_option my_long_options[] =
@@ -321,10 +320,16 @@ static struct my_option my_long_options[] =
     (uchar**)&opt_ignore_control_file, 0, 0, GET_BOOL, NO_ARG,
     0, 0, 0, 0, 0, 0},
   {"keys-used", 'k',
-   "Tell Aria to update only some specific keys. # is a bit mask of which keys to use. This can be used to get faster inserts.",
+   "Tell Aria to update only some specific keys. # is a bit mask of which keys to use. This can be used to get faster inserts. See also keys-active",
    &check_param.keys_in_use,
    &check_param.keys_in_use,
    0, GET_ULL, REQUIRED_ARG, -1, 0, 0, 0, 0, 0},
+  {"keys-active", OPT_ACTIVE_KEYS,
+   "Threat all not listed keys as disabled. If used with repair, the keys "
+   "will be disabled permanently. The argument is a list of key numbers, "
+   "starting from 1, separated by ','. "
+   "keys-active and keys-used are two ways to do the same thing",
+   0, 0, 0, GET_STR, REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
   {"datadir", 'h',
    "Path for control file (and logs if --logdir not used).",
    (char**) &maria_data_root, 0, 0, GET_STR, REQUIRED_ARG,
@@ -419,24 +424,24 @@ static struct my_option my_long_options[] =
     "Size of page buffer. Used by --safe-repair",
     &check_param.use_buffers, &check_param.use_buffers, 0,
     GET_ULONG, REQUIRED_ARG, PAGE_BUFFER_INIT, 1024L*1024L,
-    SIZE_T_MAX, (long) MALLOC_OVERHEAD, (long) IO_SIZE, 0},
+    SIZE_T_MAX, 0, (long) IO_SIZE, 0},
   { "read_buffer_size", OPT_READ_BUFFER_SIZE,
     "Read buffer size for sequential reads during scanning",
     &check_param.read_buffer_length,
     &check_param.read_buffer_length, 0, GET_ULONG, REQUIRED_ARG,
     (long) READ_BUFFER_INIT, (long) MALLOC_OVERHEAD,
-    ~0ULL, (long) MALLOC_OVERHEAD, (long) 1L, 0},
+    ~0ULL, 0, (long) 1L, 0},
   { "write_buffer_size", OPT_WRITE_BUFFER_SIZE,
     "Write buffer size for sequential writes during repair of fixed size or dynamic size rows",
     &check_param.write_buffer_length,
     &check_param.write_buffer_length, 0, GET_ULONG, REQUIRED_ARG,
     (long) READ_BUFFER_INIT, (long) MALLOC_OVERHEAD,
-    ~0UL, (long) MALLOC_OVERHEAD, (long) 1L, 0},
+    ~0UL, 0, (long) 1L, 0},
   { "sort_buffer_size", OPT_SORT_BUFFER_SIZE,
     "Size of sort buffer. Used by --recover",
     &check_param.orig_sort_buffer_length,
     &check_param.orig_sort_buffer_length, 0, GET_ULL, REQUIRED_ARG,
-    SORT_BUFFER_INIT, MIN_SORT_BUFFER, SIZE_T_MAX, MALLOC_OVERHEAD, 1L, 0},
+    SORT_BUFFER_INIT, MARIA_MIN_SORT_MEMORY, SIZE_T_MAX/10, 0, 1L, 0},
   { "sort_key_blocks", OPT_SORT_KEY_BLOCKS,
     "Internal buffer for sorting keys; Don't touch :)",
     &check_param.sort_key_blocks,
@@ -456,7 +461,7 @@ static struct my_option my_long_options[] =
     REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
   { "stats_method", OPT_STATS_METHOD,
     "Specifies how index statistics collection code should treat NULLs. "
-    "Possible values of name are \"nulls_unequal\" (default behavior for 4.1/5.0), "
+    "Possible values of name are \"nulls_unequal\" (default behavior for MySQL 4.1/5.0), "
     "\"nulls_equal\" (emulate 4.0 behavior), and \"nulls_ignored\".",
     (char**) &maria_stats_method_str, (char**) &maria_stats_method_str, 0,
     GET_STR, REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
@@ -471,13 +476,6 @@ static struct my_option my_long_options[] =
 };
 
 
-static void print_version(void)
-{
-  printf("%s  Ver 1.3 for %s on %s\n", my_progname, SYSTEM_TYPE,
-	 MACHINE_TYPE);
-}
-
-
 static void usage(void)
 {
   print_version();
@@ -505,7 +503,7 @@ static void usage(void)
 		      maria_chk very silent.\n\
   -t, --tmpdir=path   Path for temporary files. Multiple paths can be\n\
                       specified, separated by ");
-#if defined( __WIN__) || defined(__NETWARE__)
+#if defined( _WIN32)
    printf("semicolon (;)");
 #else
    printf("colon (:)");
@@ -554,7 +552,7 @@ Recover (repair)/ options (When using '--recover' or '--safe-recover'):\n\
                       file when it's full).\n\
  --create-missing-keys\n\
                       Create missing keys. This assumes that the data\n\
-                      file is correct and that the the number of rows stored\n\
+                      file is correct and that the number of rows stored\n\
                       in the index file is correct. Enables --quick.\n\
   -e, --extend-check  Try to recover every possible row from the data file\n\
 		      Normally this will also find a lot of garbage rows;\n\
@@ -567,6 +565,11 @@ Recover (repair)/ options (When using '--recover' or '--safe-recover'):\n\
   -k, --keys-used=#   Tell Aria to update only some specific keys. # is a\n\
 	              bit mask of which keys to use. This can be used to\n\
 		      get faster inserts.\n\
+  --keys-active\n\
+   Treat all not listed keys as disabled. If used with repair, the keys\n\
+   will be disabled permanently. The argument is a list of key numbers,\n\
+   starting from 1, separated by ','\n\
+   keys-active and keys-used are two ways to do the same thing\n\
   --max-record-length=#\n\
                       Skip rows bigger than this if aria_chk can't allocate\n\
 		      memory to hold it.\n\
@@ -638,9 +641,7 @@ Recover (repair)/ options (When using '--recover' or '--safe-recover'):\n\
 
 const char *maria_stats_method_names[] = {"nulls_unequal", "nulls_equal",
                                            "nulls_ignored", NullS};
-TYPELIB maria_stats_method_typelib= {
-  array_elements(maria_stats_method_names) - 1, "",
-  maria_stats_method_names, NULL};
+TYPELIB maria_stats_method_typelib= CREATE_TYPELIB_FOR(maria_stats_method_names);
 
 	 /* Read options */
 
@@ -751,6 +752,32 @@ get_one_option(const struct my_option *opt,
   case 'k':
     check_param.keys_in_use= (ulonglong) strtoll(argument, NULL, 10);
     break;
+  case OPT_ACTIVE_KEYS:
+    if (argument == disabled_my_option)
+      check_param.keys_in_use= ~0LL;
+    else
+    {
+      const char *start;
+      char *end, *str_end= strend(argument);
+      check_param.keys_in_use= 0;
+      for (start= argument; *start ; start= end)
+      {
+        int error;
+        longlong key;
+        end= str_end;
+        key= my_strtoll10(start, &end, &error);
+        if (error || key > 64 || (*end && *end != ','))
+        {
+          fprintf(stderr, "Wrong argument to active-keys. Expected a list of "
+                  "numbers like 1,2,3,4\n");
+          exit(1);                              /* Change to my_exit after merge */
+        }
+        check_param.keys_in_use|= 1LL << (key-1);
+        if (*end == ',')
+          end++;
+      }
+    }
+    break;
   case 'm':
     if (argument == disabled_my_option)
       check_param.testflag&= ~T_MEDIUM;
@@ -986,7 +1013,7 @@ static void get_options(register int *argc,register char ***argv)
 
   if (set_collation_name)
     if (!(set_collation= get_charset_by_name(set_collation_name,
-                                             MYF(MY_WME))))
+                                             MYF(MY_UTF8_IS_UTF8MB3 | MY_WME))))
       my_exit(1);
 
   if (maria_data_root != default_log_dir && opt_log_dir == default_log_dir)
@@ -1033,7 +1060,9 @@ static int maria_chk(HA_CHECK *param, char *filename)
   if (!(info=maria_open(filename,
                         (param->testflag & (T_DESCRIPT | T_READONLY)) ?
                         O_RDONLY : O_RDWR,
-                        HA_OPEN_FOR_REPAIR |
+                        HA_OPEN_FOR_REPAIR | HA_OPEN_FORCE_MODE |
+                        ((param->testflag & T_QUICK) ?
+                         HA_OPEN_DATA_READONLY : 0) |
                         ((param->testflag & T_WAIT_FOREVER) ?
                          HA_OPEN_WAIT_IF_LOCKED :
                          (param->testflag & T_DESCRIPT) ?
@@ -1161,6 +1190,13 @@ static int maria_chk(HA_CHECK *param, char *filename)
       DBUG_RETURN(0);
     }
   }
+
+  /* Don't allow disable of active auto_increment keys for repair */
+  if (share->base.auto_key &&
+      (share->state.key_map & (1LL << share->base.auto_key)) &&
+      param->testflag & (T_REP_ANY | T_SORT_RECORDS | T_SORT_INDEX))
+    check_param.keys_in_use|= (1LL << share->base.auto_key);
+
   if ((param->testflag & (T_REP_ANY | T_STATISTICS |
 			  T_SORT_RECORDS | T_SORT_INDEX)) &&
       (((param->testflag & T_UNPACK) &&
@@ -1344,7 +1380,8 @@ static int maria_chk(HA_CHECK *param, char *filename)
         */
         my_bool update_index=1;
         for (key=0 ; key < share->base.keys; key++)
-          if (share->keyinfo[key].flag & (HA_BINARY_PACK_KEY|HA_FULLTEXT))
+          if (share->keyinfo[key].flag & HA_BINARY_PACK_KEY ||
+              share->keyinfo[key].key_alg == HA_KEY_ALG_FULLTEXT)
             update_index=0;
 
         error=maria_sort_records(param,info,filename,param->opt_sort_key,
@@ -1369,6 +1406,7 @@ static int maria_chk(HA_CHECK *param, char *filename)
       DBUG_PRINT("info", ("Resetting crashed state"));
       share->state.changed&= ~(STATE_CHANGED | STATE_CRASHED_FLAGS |
                                STATE_IN_REPAIR);
+      share->crash_error= 0;
     }
     else
       maria_mark_crashed(info);
@@ -1433,6 +1471,7 @@ static int maria_chk(HA_CHECK *param, char *filename)
       DBUG_PRINT("info", ("Resetting crashed state"));
       share->state.changed&= ~(STATE_CHANGED | STATE_CRASHED_FLAGS |
                                STATE_IN_REPAIR);
+      share->crash_error= 0;
     }
     else if (!maria_is_crashed(info) &&
              (param->testflag & T_UPDATE_STATE))
@@ -1592,8 +1631,14 @@ static void descript(HA_CHECK *param, register MARIA_HA *info, char *name)
     }
     compile_time_assert((MY_UUID_STRING_LENGTH + 1) <= sizeof(buff));
     buff[MY_UUID_STRING_LENGTH]= 0;
-    my_uuid2str(share->base.uuid, buff);
+    my_uuid2str(share->base.uuid, buff, 1);
     printf("UUID:                %s\n", buff);
+    if (ma_control_file_inited() &&
+        memcmp(share->base.uuid, maria_uuid, MY_UUID_SIZE))
+      printf("Warning: File UUID not match control file UUID! "
+             "File is probably moved\n"
+             "It will be updated to new system on first usage if zerofill is "
+             "not done\n");
     pos=buff;
     if (share->state.changed & STATE_CRASHED)
       strmov(buff, share->state.changed & STATE_CRASHED_ON_REPAIR ?
@@ -1618,6 +1663,8 @@ static void descript(HA_CHECK *param, register MARIA_HA *info, char *name)
 	pos=strmov(pos,"sorted index pages,");
       if (!(share->state.changed & STATE_NOT_ZEROFILLED))
 	pos=strmov(pos,"zerofilled,");
+      if (test_all_bits(share->state.changed, (STATE_NOT_ZEROFILLED | STATE_HAS_LSN)))
+        pos=strmov(pos,"has_lsn,");
       if (!(share->state.changed & STATE_NOT_MOVABLE))
 	pos=strmov(pos,"movable,");
       if (have_control_file && (share->state.changed & STATE_MOVED))
@@ -1690,7 +1737,7 @@ static void descript(HA_CHECK *param, register MARIA_HA *info, char *name)
   {
     keyseg=keyinfo->seg;
     if (keyinfo->flag & HA_NOSAME) text="unique ";
-    else if (keyinfo->flag & HA_FULLTEXT) text="fulltext ";
+    else if (keyinfo->key_alg == HA_KEY_ALG_FULLTEXT) text="fulltext ";
     else text="multip.";
 
     pos=buff;
@@ -1876,7 +1923,7 @@ static int maria_sort_records(HA_CHECK *param,
     param->error_printed=0;
     DBUG_RETURN(0);				/* Nothing to do */
   }
-  if (keyinfo->flag & HA_FULLTEXT)
+  if (keyinfo->key_alg == HA_KEY_ALG_FULLTEXT)
   {
     _ma_check_print_warning(param,"Can't sort table '%s' on FULLTEXT key %d",
                             name,sort_key+1);
